/*
 * poppassd_pam.c
 *
 * A Eudora and NUPOP change password server (PAM version).
 *
 * Copyright by Wade Turland <w.turland@uws.edu.au>, Information Technology
 * Directorate, University of Western Sydney
 *
 * THE UNIVERSITY OF WESTERN SYDNEY DISCLAIMS ALL WARRANTIES, EXPRESS OR
 * IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE,
 * AND IN NO EVENT SHALL THE UNIVERSITY OF WESTERN SYDNEY BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Portions copyright by University of Washington.
 *
 * Adapted from the original poppassd work by: John Norstad
 * <j-norstad@nwu.edu>, Academic Computing and Network Services, Northwestern
 * University.
 *
 * Based on earlier versions by Roy Smith <roy@nyu.edu> and Daniel L. Leavitt
 * <dll.mitre.org>.
 * 
 * Note that unencrypted passwords are transmitted over the network.  If this
 * bothers you, think hard about whether you want to implement the password
 * changing feature.  On the other hand, it's no worse than what happens when
 * you run /bin/passwd while connected via telnet or rlogin.  Well, maybe it
 * is, since the use of a dedicated port makes it slightly easier for a network
 * snooper to snarf passwords off the wire.
 *
 * NOTE: In addition to the security issue outlined in the above paragraph, you
 * should be aware that this program is going to be run as root by ordinary
 * users and it mucks around with the password file.  This should set alarms
 * off in your head.  I think I've devised a pretty foolproof way to ensure
 * that security is maintained, but I'm no security expert and you would be a
 * fool to install this without first reading the code and ensuring yourself
 * that what I consider safe is good enough for you.  If something goes wrong,
 * it's your fault, not mine.
 *
 * The front-end code (which talks to the client) is directly descended from
 * Leavitt's original version.
 *        
 * Should be owned by root, and executable only by root.  It can be started
 * with an entry in /etc/inetd.conf such as the following:
 *
 * poppassd stream tcp nowait root /usr/local/bin/poppassd poppassd
 * 
 * and in /etc/services:
 * 
 * poppassd    106/tcp
 *
 * Create entries in /etc/pam.conf like this (Solaris 2.x):
 * poppassd password required /usr/lib/security/$ISA/pam_unix.so.1
 * poppassd password optional /opt/pam_ldap-122/lib/security/pam_ldap.so
 *
 * Logs to the local2 facility. Should have an entry in /etc/syslog.conf like
 * the following:
 *
 * local2.err    /var/adm/poppassd-log
 */
 
/* Modification history.
 *
 * 10/10/01. Version 1.0. w.turland@uws.edu.au
 * Converted old pty poppassd to pam version
 */

/* Steve Dorner's description of the simple protocol:
 *
 * The server's responses should be like an FTP server's responses; 
 * 1xx for in progress, 2xx for success, 3xx for more information
 * needed, 4xx for temporary failure, and 5xx for permanent failure.  
 * Putting it all together, here's a sample conversation:
 *
 *   S: 200 hello\r\n
 *   E: user yourloginname\r\n
 *   S: 300 please send your password now\r\n
 *   E: pass yourcurrentpassword\r\n
 *   S: 200 My, that was tasty\r\n
 *   E: newpass yournewpassword\r\n
 *   S: 200 Happy to oblige\r\n
 *   E: quit\r\n
 *   S: 200 Bye-bye\r\n
 *   S: <closes connection>
 *   E: <closes connection>
 */
 
#define VERSION "1.0"

#define BUFSIZE PAM_MAX_MSG_SIZE

#include <pwd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <strings.h>
#include <syslog.h>
#include <security/pam_appl.h>

/*
 * Prototypes
 */

int main (int argc, char *argv[]);
void WriteToClient (char *fmt, ...);
void ReadFromClient (char *line);
int chPass (char *user, char *pass, char *newpass);
static int chPass_conv (int num_msg,struct pam_message **msg,
             struct pam_response **resp,void *appdata_ptr);
int chPass (char *user, char *pass, char *newpass);


struct chPass_cred {
    char *uname;            /* user name */
    char *tok;              /* password */
    char *newtok;           /* new password */
};

int
main (int argc, char *argv[]) {
    struct passwd *pw;
    char user[BUFSIZE] = "";
    char line[BUFSIZE] = "";
    char oldpass[BUFSIZE] = "";
    char newpass[BUFSIZE] = "";

    openlog ("poppassd", LOG_PID, LOG_LOCAL2);

    gethostname(line, sizeof (line));
    WriteToClient ("200 %s poppassd v%s hello, who are you?", line, VERSION);

    ReadFromClient (line);
    sscanf (line, "user %s", user) ;
    if (strlen (user) == 0) {
        WriteToClient ("500 Username required.");
        return(1);
    }

    WriteToClient ("200 your password please.");
    ReadFromClient (line);
    sscanf (line, "pass %s", oldpass) ;

    if (strlen (oldpass) == 0) {
        WriteToClient ("500 Password required.");
        return(1);
    }

    WriteToClient ("200 your new password please.");
    ReadFromClient (line);
    sscanf (line, "newpass %s", newpass);

    /* new pass required */
    if (strlen (newpass) == 0) {
        WriteToClient ("500 New password required.");
        return(1);
    }

    if ((pw = getpwnam (user)) == NULL) {
        WriteToClient ("500 Invalid user or password");
        return(1);
    }

    /* true if failure */
    if (chPass(pw->pw_name, oldpass, newpass)) {
        syslog (LOG_ERR, "failed password change for %s", user);
        return(1);
    }

    syslog (LOG_ERR, "password changed for %s", user);
    WriteToClient ("200 Password changed, thank-you.");

    ReadFromClient (line);
    if (strncmp(line, "quit", 4) != 0) {
        WriteToClient("500 Quit required.");
        return (1);
    }

    WriteToClient("200 Bye.");
    return (0);
}

void
WriteToClient (char *fmt, ...) {
    va_list ap;
    
    va_start (ap, fmt);
    vfprintf (stdout, fmt, ap);
    fputs ("\r\n", stdout );
    fflush (stdout);
    va_end (ap);
}

void
ReadFromClient (char *line) {
    char *sp;
    int i;

    fgets (line, BUFSIZE, stdin);
    if ((sp = strchr(line, '\n')) != NULL) *sp = '\0'; 
    if ((sp = strchr(line, '\r')) != NULL) *sp = '\0'; 

    /* convert initial keyword on line to lower case. */
    for (sp = line; isalpha(*sp); sp++)
        *sp = tolower(*sp);
}

/* PAM conversation function
 * Accepts: number of messages
 *        vector of messages
 *        pointer to response return
 *        application data
 * Returns: PAM_SUCCESS if OK, response vector filled in, else PAM_CONV_ERR
 */
static int
chPass_conv (int num_msg,struct pam_message **msg,
             struct pam_response **resp,void *appdata_ptr) {
    int i;
    char *j;
    struct chPass_cred *cred = (struct chPass_cred *) appdata_ptr;
    struct pam_response *reply = calloc(sizeof (struct pam_response), num_msg);
    for (i = 0; i < num_msg; i++) {
        switch (msg[i]->msg_style) {
            case PAM_PROMPT_ECHO_ON:    /* assume want user name */
                reply[i].resp_retcode = PAM_SUCCESS;
                reply[i].resp = strdup (cred->uname);
                break;
            case PAM_PROMPT_ECHO_OFF:    /* assume want password */
                reply[i].resp = NULL;
                for (j = (char *)msg[i]->msg; *j != (char) NULL; j++) {
                    if (strncasecmp (j,"new",3) == 0) {
                        reply[i].resp_retcode = PAM_SUCCESS;
                        reply[i].resp = strdup (cred->newtok);
                        break;
                    }
                }
                if (reply[i].resp == NULL) {
                    reply[i].resp_retcode = PAM_SUCCESS;
                    reply[i].resp = strdup (cred->tok);
                }
                break;
            case PAM_TEXT_INFO:
            case PAM_ERROR_MSG:
                WriteToClient("100 %s",msg[i]->msg);
                reply[i].resp_retcode = PAM_SUCCESS;
                reply[i].resp = NULL;
                break;
            default:            /* unknown message style */
                return PAM_CONV_ERR;
        }
    }
    *resp = reply;
    return PAM_SUCCESS;
}

int
chPass (char *user, char *pass, char *newpass) {
    pam_handle_t *hdl;
    struct pam_conv conv;
    struct chPass_cred cred;
    int rc;
    conv.conv = (void *) &chPass_conv;
    conv.appdata_ptr = &cred;
    cred.uname = user;
    cred.tok = pass;
    cred.newtok = newpass;
    rc = pam_start ("poppassd", user, &conv, &hdl);
    if (rc != PAM_SUCCESS) {
        WriteToClient("500 %s",pam_strerror(hdl, rc));
        return (1);
    }
    rc = pam_chauthtok(hdl, PAM_SUCCESS);
    if (rc != PAM_SUCCESS) {
        WriteToClient("500 %s",pam_strerror(hdl, rc));
        pam_end (hdl, PAM_SUCCESS); /* clean up */
        return (1);
    }
    pam_end (hdl, PAM_SUCCESS); /* return success */
    return (0);
}
